/*
 * 3-D Helixpaint. Inspired by Electropaint.
 */

#include <stdlib.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkgl.h>
#include <GL/gl.h>

#include "helix.h"
#include "midi-source.h"
#include "midi-adjustment.h"

static HelixAttribs  attr_zero = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, {0.0, 0.0, 0.0}, 0 };

static HelixSlotParams helix_defaults[HELIX_SLOT_LAST] =
{
   /*  slot, lower, upper, value, midi_ctrl, cyclic, name  */
  { HELIX_ARMLEN_ABS,      0.0,   1.0,   0.5,  16, FALSE, "Armlength (abs)" },
  { HELIX_ARMLEN_REL,      0.0,   0.3,   0.0,   0, FALSE, "Armlenght (rel)" },
  { HELIX_ARMLEN_AMP,      0.0,   0.3,   0.0,   0, FALSE, "Armlenght (amp)" },
  { HELIX_HEIGHT_ABS,     -0.5,   0.5,   0.0,  17, FALSE, "Height (abs)"    },
  { HELIX_HEIGHT_REL,      0.0,  0.05,   0.02, 18, FALSE, "Height (rel)"    },
  { HELIX_HEIGHT_AMP,      0.0,   0.3,   0.0,   0, FALSE, "Height (amp)"    },
  { HELIX_SIZE_ABS,        0.0,   0.3,   0.08, 19, FALSE, "Size (abs)"      },
  { HELIX_SIZE_REL,       -0.1,   0.1,   0.0,   0, FALSE, "Size (rel)"      },
  { HELIX_SIZE_AMP,        0.0,   0.3,   0.0,   0, FALSE, "Size (amp)"      },
  { HELIX_ANGLE_ABS,       0.0, 180.0,   0.0,   0, FALSE, "Angle (abs)"     },
  { HELIX_ANGLE_REL,       0.0, 180.0,  20.0,  20, FALSE, "Angle (rel)"     },
  { HELIX_ANGLE_AMP,       0.0, 180.0,   0.0,   0, FALSE, "Angle (amp)"     },
  { HELIX_ROLL_ABS,     -180.0, 180.0,   0.0,  21, TRUE,  "Roll (abs)"      },
  { HELIX_ROLL_REL,     -180.0, 180.0,   0.0,   0, TRUE,  "Roll (rel)"      },
  { HELIX_ROLL_AMP,        0.0, 180.0,   0.0,   0, FALSE, "Roll (amp)"      },
  { HELIX_WAVE_ABS,     -180.0, 180.0,   0.0,  22, TRUE,  "Wave (abs)"      },
  { HELIX_WAVE_REL,     -180.0, 180.0,   0.0,   0, TRUE,  "Wave (rel)"      },
  { HELIX_WAVE_AMP,        0.0, 180.0,   0.0,   0, FALSE, "Wave (amp)"      },
  { HELIX_WAG_ABS,      -180.0, 180.0,   0.0,  23, TRUE,  "Wag (abs)"       },
  { HELIX_WAG_REL,      -180.0, 180.0,   0.0,   0, TRUE,  "Wag (rel)"       },
  { HELIX_WAG_AMP,         0.0, 180.0,   0.0,   0, FALSE, "Wag (amp)"       },
  { HELIX_L_ABS,           0.0, 100.0,  70.0,  27, FALSE, "L* (abs)"        },
  { HELIX_L_REL,          -2.5,   2.5,  -1.5,  28, FALSE, "L* (rel)"        },
  { HELIX_L_AMP,           0.0,  50.0,   0.0,   0, FALSE, "L* (amp)"        },
  { HELIX_A_ABS,        -100.0, 100.0,   0.0,  29, FALSE, "a* (abs)"        },
  { HELIX_A_REL,          -5.0,   5.0,   0.0,  30, FALSE, "a* (rel)"        },
  { HELIX_A_AMP,           0.0,  50.0,   0.0,   0, FALSE, "a* (amp)"        },
  { HELIX_B_ABS,        -100.0, 100.0, 100.0,  31, FALSE, "b* (abs)"        },
  { HELIX_B_REL,          -5.0,   5.0,  -3.7,  33, FALSE, "b* (rel)"        },
  { HELIX_B_AMP,           0.0,  50.0,   0.0,   0, FALSE, "b* (amp)"        },
  { HELIX_OPACITY_ABS,     0.0,   1.0,   1.0,  24, FALSE, "Opacity (abs)"   },
  { HELIX_OPACITY_REL,    -1.0,   1.0,   0.0,   0, FALSE, "Opacity (rel)"   },
  { HELIX_OPACITY_AMP,     0.0,   1.0,   0.0,   0, FALSE, "Opacity (amp)"   },
  { HELIX_N_VERTICES,      3.0,   16.0,  4.0,   0, FALSE, "n Vertices"      },
  { HELIX_VIEW_ROT_X,   -180.0, 180.0, -50.0,   1, TRUE,  "X-Angle"         },
  { HELIX_VIEW_ROT_Y,   -180.0, 180.0,   0.0,   2, TRUE,  "Y-Angle"         },
  { HELIX_VIEW_ROT_Z,   -180.0, 180.0,   0.0,   6, TRUE,  "Z-Angle"         },
  { HELIX_ROTATION_REL,   -1.0,   1.0,   0.04, 25, FALSE, "Rotation speed"  },
  { HELIX_FADEOUT,        0.0,    1.0,   1.0,   0, FALSE, "Fadeout"         }
};

static void
helix_elements_configure (HelixWindow *hwin)
{
  gint i;

  hwin->master_element->absolute = attr_zero;
  hwin->master_element->relative = attr_zero;
  hwin->master_element->absolute.vertices = 4;

  hwin->oscillation_params = attr_zero;

  for (i = 0; i < hwin->nelements; i++)
    {
      HelixElement *elem = &hwin->elements[i];

      elem->absolute = hwin->master_element->absolute;
      elem->relative = hwin->master_element->relative;
    }
}


static void
helix_elements_propagate (HelixWindow *hwin)
{
  HelixElement *elem, *prev;
  gint i;

  for (i = hwin->nelements - 1; i >= 0; i--)
    {
      elem = &hwin->elements[i];

      if (i > 0)
        prev = &hwin->elements[i-1];
      else
        prev = hwin->master_element;

      /* copy behaviour from previous element */
      elem->absolute = prev->absolute;
      elem->relative = prev->relative;
    }
}


static void
helix_attribs_mix (gdouble        prev_factor,
                   HelixAttribs  *prev,
                   gdouble        this_factor,
                   HelixAttribs  *this,
                   HelixAttribs  *accu)
{
  gdouble d;

  accu->armlen  = prev_factor * prev->armlen  + this_factor * this->armlen;
  accu->height  = prev_factor * prev->height  + this_factor * this->height;
  accu->size    = prev_factor * prev->size    + this_factor * this->size;
  accu->opacity = prev_factor * prev->opacity + this_factor * this->opacity;

  d = this->wave - prev->wave;
  accu->wave    = (prev_factor * (d > 180  ? prev->wave + 360 : prev->wave) +
                   this_factor * (d < -180 ? this->wave + 360 : this->wave));
  d = this->roll - prev->roll;
  accu->roll    = (prev_factor * (d > 180  ? prev->roll + 360 : prev->roll) +
                   this_factor * (d < -180 ? this->roll + 360 : this->roll));
  d = this->wag - prev->wag;
  accu->wag     = (prev_factor * (d > 180  ? prev->wag + 360 : prev->wag) +
                   this_factor * (d < -180 ? this->wag + 360 : this->wag));

  d = this->angle - prev->angle;
  accu->angle   = (prev_factor * (d > 180  ? prev->angle + 360 : prev->angle) +
                   this_factor * (d < -180 ? this->angle + 360 : this->angle));

  accu->vertices = (prev_factor * prev->vertices +
                    this_factor * this->vertices);

  accu->Lab[0]  = prev_factor * prev->Lab[0]  + this_factor * this->Lab[0];
  accu->Lab[1]  = prev_factor * prev->Lab[1]  + this_factor * this->Lab[1];
  accu->Lab[2]  = prev_factor * prev->Lab[2]  + this_factor * this->Lab[2];
}


static int
helix_z_order_compare (const void *a,
                       const void *b)
{
  HelixIndexZ *a1 = (HelixIndexZ *) a;
  HelixIndexZ *b1 = (HelixIndexZ *) b;

  if (a1->z_depth < b1->z_depth)
    return -1;
  else if (a1->z_depth > b1->z_depth)
    return 1;
  else
    return 0;
}


/* draw a frame */
static gboolean
helix_frame_draw (GtkWidget      *widget,
                  GdkEventExpose *event,
                  gpointer        data)
{
  HelixWindow  *hwin = (HelixWindow *) data;
  HelixElement *prev;
  GLfloat color [4]     = { 0.0, 0.0, 0.0, 0.5 };
  GLfloat specular [4]  = { 0.5, 0.5, 0.5, 1.0 };
  GLfloat shininess [1] = { 70.0 };
  gint j;
  gdouble angle, offset, eo_reminder;
  static gdouble total_height = 0;
  HelixIndexZ  *ordering_z = g_new0 (HelixIndexZ, hwin->nelements);
  HelixAttribs accu_rel = attr_zero;
  gdouble  rot_x, rot_y, rot_z;

  hwin->rotation_z += gtk_adjustment_get_value (hwin->adjustments[HELIX_ROTATION_REL]) * 360 / hwin->fps;

  rot_x = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_X]);
  rot_y = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_Y]);
  rot_z = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_Z]);

  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glPushMatrix ();
  glRotatef (rot_x, 1.0, 0.0, 0.0);
  glRotatef (rot_y, 0.0, 1.0, 0.0);
  glRotatef (hwin->rotation_z + rot_z, 0.0, 0.0, 1.0);
  glTranslatef (0.0, 0.0, - total_height / 2.0);

  hwin->master_element->absolute.armlen  = gtk_adjustment_get_value (hwin->adjustments[HELIX_ARMLEN_ABS]);
  hwin->master_element->relative.armlen  = gtk_adjustment_get_value (hwin->adjustments[HELIX_ARMLEN_REL]);
  hwin->oscillation_params.armlen        = gtk_adjustment_get_value (hwin->adjustments[HELIX_ARMLEN_AMP]);
  hwin->master_element->absolute.height  = gtk_adjustment_get_value (hwin->adjustments[HELIX_HEIGHT_ABS]);
  hwin->master_element->relative.height  = gtk_adjustment_get_value (hwin->adjustments[HELIX_HEIGHT_REL]);
  hwin->oscillation_params.height        = gtk_adjustment_get_value (hwin->adjustments[HELIX_HEIGHT_AMP]);
  hwin->master_element->absolute.size    = gtk_adjustment_get_value (hwin->adjustments[HELIX_SIZE_ABS]);
  hwin->master_element->relative.size    = gtk_adjustment_get_value (hwin->adjustments[HELIX_SIZE_REL]);
  hwin->oscillation_params.size          = gtk_adjustment_get_value (hwin->adjustments[HELIX_SIZE_AMP]);
  hwin->master_element->absolute.angle   = gtk_adjustment_get_value (hwin->adjustments[HELIX_ANGLE_ABS]);
  hwin->master_element->relative.angle   = gtk_adjustment_get_value (hwin->adjustments[HELIX_ANGLE_REL]);
  hwin->oscillation_params.angle         = gtk_adjustment_get_value (hwin->adjustments[HELIX_ANGLE_AMP]);
  hwin->master_element->absolute.roll    = gtk_adjustment_get_value (hwin->adjustments[HELIX_ROLL_ABS]);
  hwin->master_element->relative.roll    = gtk_adjustment_get_value (hwin->adjustments[HELIX_ROLL_REL]);
  hwin->oscillation_params.roll          = gtk_adjustment_get_value (hwin->adjustments[HELIX_ROLL_AMP]);
  hwin->master_element->absolute.wave    = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAVE_ABS]);
  hwin->master_element->relative.wave    = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAVE_REL]);
  hwin->oscillation_params.wave          = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAVE_AMP]);
  hwin->master_element->absolute.wag     = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAG_ABS]);
  hwin->master_element->relative.wag     = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAG_REL]);
  hwin->oscillation_params.wag           = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAG_AMP]);
  hwin->master_element->absolute.opacity = gtk_adjustment_get_value (hwin->adjustments[HELIX_OPACITY_ABS]);
  hwin->master_element->relative.opacity = gtk_adjustment_get_value (hwin->adjustments[HELIX_OPACITY_REL]);
  hwin->oscillation_params.opacity       = gtk_adjustment_get_value (hwin->adjustments[HELIX_OPACITY_AMP]);
  hwin->master_element->absolute.Lab[0]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_L_ABS]);
  hwin->master_element->relative.Lab[0]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_L_REL]);
  hwin->oscillation_params.Lab[0]        = gtk_adjustment_get_value (hwin->adjustments[HELIX_L_AMP]);
  hwin->master_element->absolute.Lab[1]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_A_ABS]);
  hwin->master_element->relative.Lab[1]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_A_REL]);
  hwin->oscillation_params.Lab[1]        = gtk_adjustment_get_value (hwin->adjustments[HELIX_A_AMP]);
  hwin->master_element->absolute.Lab[2]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_B_ABS]);
  hwin->master_element->relative.Lab[2]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_B_REL]);
  hwin->oscillation_params.Lab[2]        = gtk_adjustment_get_value (hwin->adjustments[HELIX_B_AMP]);
  hwin->master_element->absolute.vertices = gtk_adjustment_get_value (hwin->adjustments[HELIX_N_VERTICES]);

  hwin->master_element->absolute.armlen  += hwin->oscillation_params.armlen * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.height  += hwin->oscillation_params.height * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.size    += hwin->oscillation_params.size * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.angle   += hwin->oscillation_params.angle * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.roll    += hwin->oscillation_params.roll * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.wave    += hwin->oscillation_params.wave * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.wag     += hwin->oscillation_params.wag * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.opacity += hwin->oscillation_params.opacity * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.Lab[0]  += hwin->oscillation_params.Lab[0] * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.Lab[1]  += hwin->oscillation_params.Lab[1] * sin (hwin->ticks + hwin->tick_fraction);
  hwin->master_element->absolute.Lab[2]  += hwin->oscillation_params.Lab[2] * sin (hwin->ticks + hwin->tick_fraction);

  if (hwin->texture && hwin->textured)
    {
      glEnable (GL_TEXTURE_2D);
      glDisable (GL_DEPTH_TEST);
    }
  else
    {
      glDisable (GL_TEXTURE_2D);
      glEnable (GL_DEPTH_TEST);
    }

  /*
   * calculate the current state of the elements
   * and store their calculated transformation matrix
   */
  prev = hwin->master_element;

  for (j = 0; j < hwin->nelements; j++)
    {
      HelixElement *elem = &hwin->elements[j];
      HelixAttribs  curr_rel;

      gdouble dt = hwin->tick_fraction;

      /* interpolate the relative and absolute attributes */
      helix_attribs_mix (1.0 - dt, &elem->relative,
                         dt,       &prev->relative,
                         &curr_rel);

      helix_attribs_mix (1.0 - dt, &elem->absolute,
                         dt,       &prev->absolute,
                         &elem->current);

      /* accumulate the current relative attributes */
      helix_attribs_mix (1.0, &accu_rel,
                         1.0, &curr_rel,
                         &accu_rel);

      /*
       * the used values are the accumulated relative plus
       * the current absolute values.
       */
      helix_attribs_mix (1.0, &elem->current,
                         1.0, &accu_rel,
                         &elem->current);

      glPushMatrix ();

      glRotated (elem->current.angle, 0.0, 0.0, 1.0);
      glTranslated (0.0, elem->current.armlen, elem->current.height);
      glRotated (elem->current.roll, 0.0, 1.0, 0.0);
      glRotated (elem->current.wave, 1.0, 0.0, 0.0);
      glRotated (elem->current.wag,  0.0, 0.0, 1.0);
      glScaled (fabs (elem->current.size),
                fabs (elem->current.size),
                fabs (elem->current.size));

      /*
       * calculate the depth of the center of a element in world coords.
       * the center is (0,0,0,1)  (homogenous), thus
       * M * (0,0,0,1) =  (m[12], m[13], m[14], m[15])
       * and unhomogenizing gives z = m[14]/m[15].
       */
      glGetDoublev (GL_MODELVIEW_MATRIX, elem->transform);
      elem->z_depth = elem->transform[14] / elem->transform[15];

      ordering_z[j].index = j;
      ordering_z[j].z_depth = elem->z_depth;

      glPopMatrix ();

      prev = elem;
    }

  /*
   * sort the indicies of the elements according to their distance
   * to the viewer
   */

  qsort (ordering_z, hwin->nelements,
         sizeof (HelixIndexZ), helix_z_order_compare);

  /*
   * finally draw them to the screen in the correct order.
   */


  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  for (j = 0; j < hwin->nelements; j++)
    {
      HelixElement *elem = &hwin->elements[ordering_z[j].index];
      gdouble depth, opacity;
      gdouble X, Y, Z, R, G, B;

      glLoadMatrixd (elem->transform);

      depth = CLAMP ((elem->z_depth + 3.5), 0.0, 1.0);
      opacity = CLAMP (elem->current.opacity, 0.0, 1.0);

      if (hwin->texture && hwin->textured)
        {
          color[0] = color[1] = color[2] = 1.0;
          color[3] = opacity;
        }
      else
        {
          /*  as seen on http://www.easyrgb.com/math.php  */
          /*  L*a*b*  -->  XYZ  */

          Y = (elem->current.Lab[0] + 16) / 116;
          X = elem->current.Lab[1] / 500 + Y;
          Z = Y - elem->current.Lab[2] / 200;

          if (pow (Y, 3) > 0.008856)
            Y = pow (Y, 3);
          else
            Y = (Y - 16. / 116) / 7.787;

          if (pow (X, 3) > 0.008856)
            X = pow (X, 3);
          else
            X = (X - 16. / 116) / 7.787;

          if (pow (Z, 3) > 0.008856)
            Z = pow (Z, 3);
          else
            Z = (Z - 16. / 116) / 7.787;

          /* Observer= 2, Illuminant= D65  */
          X *=  95.047 / 100;
          Y *= 100.000 / 100;
          Z *= 108.883 / 100;

          /*  XYZ --> RGB  */

          R = X *  3.2406 + Y * -1.5372 + Z * -0.4986;
          G = X * -0.9689 + Y *  1.8758 + Z *  0.0415;
          B = X *  0.0557 + Y * -0.2040 + Z *  1.0570;

          if (R > 0.0031308)
            R = 1.055 * pow (R, 1.0 / 2.4) - 0.055;
          else
            R = 12.92 * R;

          if (G > 0.0031308)
            G = 1.055 * pow (G, 1.0 / 2.4) - 0.055;
          else
            G = 12.92 * G;

          if (B > 0.0031308)
            B = 1.055 * pow (B, 1.0 / 2.4) - 0.055;
          else
            B = 12.92 * B;

          color[0] = CLAMP (R, 0.0, 1.0);
          color[1] = CLAMP (G, 0.0, 1.0);
          color[2] = CLAMP (B, 0.0, 1.0);
          color[3] = opacity;
        }

      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

      offset = fmod (elem->current.vertices - 1, 2.0) * G_PI / elem->current.vertices;
      angle = 2 * G_PI - offset;

      eo_reminder = pow (fmod (elem->current.vertices, 2.0), 0.4);
      glBegin(GL_POLYGON);

      /* add extra vertex for transition from an even to an odd number of vertices. */

      if (eo_reminder < 1.0)
        glVertex3f (0.0, eo_reminder + cos (angle) * (1 - eo_reminder), 0.0);

      /* clockwise to make the normal point into the correct direction */
      while (angle > 0)
        {
          gdouble x, y;

          x = sin (angle);
          y = cos (angle);

          if (hwin->texture && hwin->textured)
            glTexCoord2f (0.5 + x / 2, 0.5 + y / 2);

          glVertex3f (x, y, 0.0);
          angle -= (2.0 * G_PI / elem->current.vertices);
        }
      glEnd ();

#if 1
      color[0] = color[1] = color[2] = 1.0;
      color[3] = 1.0;

      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
      glLineWidth (1.0);

      angle = 2 * G_PI - offset;

      glBegin(GL_LINE_LOOP);
      /* clockwise to make the normal point into the correct direction */

      if (eo_reminder < 1.0)
        glVertex3f (0.0, eo_reminder + cos (angle) * (1 - eo_reminder), 0.0);

      while (angle > 0)
        {
          glVertex3f (sin (angle), cos (angle), 0.0);
          angle -= (2.0 * G_PI / elem->current.vertices);
        }
      glEnd ();
#endif
    }

  glPopMatrix ();

  total_height = accu_rel.height;

  g_free (ordering_z);

  return TRUE;
}


/* draw a stereo frame */
static gboolean
helix_frame_draw_stereo (GtkWidget      *widget,
                         GdkEventExpose *event,
                         gpointer        data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  HelixWindow *hwin = (HelixWindow *) data;
  gint width = widget->allocation.width;
  gint height = widget->allocation.height;

  GLfloat h = (GLfloat) (height) / (GLfloat) (width);
  GLfloat delta = 0.0 * 0.2;

  if (delta > 0.001)
    {
      width = width / 2;
      h = (GLfloat) (height) / (GLfloat) (width);
    }
  else
    {
      delta = 0.0;
    }

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

#if 0
  /* fade old picture */
    {
      GLfloat color[4] = { 0.0, 0.0, 0.0, 1.0 };

      color[3] = gtk_adjustment_get_value (hwin->adjustments[HELIX_FADEOUT]);

      color[3] *= color[3];

      glDisable (GL_LIGHT0);
      glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
      glClear (GL_DEPTH_BUFFER_BIT);
      glRectf (-5, -5, 5, 5);
      glClear (GL_DEPTH_BUFFER_BIT);
      glEnable (GL_LIGHT0);
    }
#else
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#endif

  glViewport (0, 0, width, height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (MIN (-.5, -.5 / h) - delta, MAX (.5, .5 / h) - delta,
             MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);
  glTranslatef (-delta, 0, 0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -3);

  helix_frame_draw (widget, event, data);

  if (delta > 0.001)
    {
      glViewport (width, 0, width, height);
      glMatrixMode (GL_PROJECTION);
      glLoadIdentity ();
      glFrustum (MIN (-.5, -.5 / h) + delta, MAX (.5, .5 / h) + delta,
                 MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);
      glTranslatef (delta, 0, 0);

      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();
      glTranslatef (0.0, 0.0, -3);

      helix_frame_draw (widget, event, data);
    }

  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
helix_frame_next (HelixWindow *hwin)
{
  gdouble ticks;

  ticks = 7 * g_timer_elapsed (hwin->timer, NULL);

  while (ticks - hwin->ticks > 1.0)
    {
      helix_elements_propagate (hwin);
      hwin->ticks += 1;
    }

  hwin->tick_fraction = ticks - hwin->ticks;

  if (hwin->visibility)
    gtk_widget_queue_draw (hwin->drawing_area);

  return TRUE;
}


/* new window size or exposure */
static gboolean
helix_configure_event (GtkWidget         *widget,
                       GdkEventConfigure *event,
                       gpointer           data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  GLfloat h = (GLfloat) (widget->allocation.height) / (GLfloat) (widget->allocation.width);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (MIN (-.5, -.5 / h), MAX (.5, .5 / h),
             MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -3);

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
helix_visibility_update (GtkWidget   *widget,
                         GdkEventAny *event,
                         gpointer     data)
{
  HelixWindow *hwin = (HelixWindow *) data;

  switch (event->type)
    {
      case GDK_MAP:
          hwin->visibility = TRUE;
          break;
      case GDK_UNMAP:
          hwin->visibility = FALSE;
          break;
      case GDK_VISIBILITY_NOTIFY:
          hwin->visibility = ((GdkEventVisibility *) event)->state != GDK_VISIBILITY_FULLY_OBSCURED;
          break;
      default:
          g_printerr ("Unknown event in helix_visibility_update!\n");
          hwin->visibility = TRUE;
          break;
    }

  return TRUE;
}


/* change view angle, exit upon ESC */
static gboolean
helix_key_press_event (GtkWidget   *widget,
                       GdkEventKey *event,
                       gpointer     data)
{
  static gboolean fullscreen = FALSE;
  static gboolean control_panel = FALSE;
  gint i;
  gdouble angle;
  HelixWindow *hwin = (HelixWindow *) data;

  switch (event->keyval)
    {
    case GDK_Up:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Down:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_Left:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Right:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_z:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Z:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_F11:
    case GDK_f:
    case GDK_F:
      if (fullscreen)
        gtk_window_unfullscreen (GTK_WINDOW (hwin->window));
      else
        gtk_window_fullscreen (GTK_WINDOW (hwin->window));
      fullscreen = !fullscreen;
      break;
    case GDK_Tab:
      if (control_panel)
        {
          gtk_widget_hide (hwin->control_panel);
        }
      else
        {
          if (!hwin->control_panel)
            hwin->control_panel = helix_create_control_panel (hwin);
          gtk_widget_show (hwin->control_panel);
        }
      control_panel = !control_panel;
      break;
    case GDK_Escape:
    case GDK_q:
    case GDK_Q:
      gtk_main_quit ();
      break;
    case GDK_T:
    case GDK_t:
      hwin->textured = ! hwin->textured;
      break;
    case GDK_R:
    case GDK_r:
      for (i = 0; i < HELIX_SLOT_LAST; i++)
        {
          g_object_set (G_OBJECT (hwin->adjustments[i]),
                        "target-value", helix_defaults[i].value, NULL);
        }
      break;
    case GDK_1:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ARMLEN_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ARMLEN_ABS]),
                    "target-value", angle + 0.01, NULL);
      break;
    case GDK_exclam:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ARMLEN_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ARMLEN_ABS]),
                    "target-value", angle - 0.01, NULL);
      break;
    case GDK_2:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_SIZE_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_SIZE_ABS]),
                    "target-value", angle + 0.01, NULL);
      break;
    case GDK_quotedbl:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_SIZE_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_SIZE_ABS]),
                    "target-value", angle - 0.01, NULL);
      break;
    case GDK_3:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ANGLE_REL]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ANGLE_REL]),
                    "target-value", angle + 2, NULL);
      break;
    case GDK_section:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ANGLE_REL]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ANGLE_REL]),
                    "target-value", angle - 2, NULL);
      break;
    case GDK_4:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ROLL_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ROLL_ABS]),
                    "target-value", angle + 5, NULL);
      break;
    case GDK_dollar:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_ROLL_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_ROLL_ABS]),
                    "target-value", angle - 5, NULL);
      break;
    case GDK_5:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_WAVE_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_WAVE_ABS]),
                    "target-value", angle + 5, NULL);
      break;
    case GDK_percent:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_WAVE_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_WAVE_ABS]),
                    "target-value", angle - 5, NULL);
      break;
    case GDK_6:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_WAG_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_WAG_ABS]),
                    "target-value", angle + 5, NULL);
      break;
    case GDK_ampersand:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_WAG_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_WAG_ABS]),
                    "target-value", angle - 5, NULL);
      break;
    case GDK_7:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_N_VERTICES]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_N_VERTICES]),
                    "target-value", angle + 1, NULL);
      break;
    case GDK_slash:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_N_VERTICES]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_N_VERTICES]),
                    "target-value", angle - 1, NULL);
      break;
    case GDK_8:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_OPACITY_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_OPACITY_ABS]),
                    "target-value", angle + 0.1, NULL);
      break;
    case GDK_parenleft:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_OPACITY_ABS]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_OPACITY_ABS]),
                    "target-value", angle - 0.1, NULL);
      break;
    default:
      return FALSE;
    }

  return TRUE;
}


static void
helix_init_gl (GtkWidget *widget,
               gpointer   data)
{
  HelixWindow *hwin = (HelixWindow *) data;
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  static GLfloat pos[4] = {5.0, 5.0, 15.0, 0.0};
  static GLfloat white[4] = {1.0, 1.0, 1.0, 1.0};

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  if (hwin->texture)
    {
      glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
      glTexImage2D (GL_TEXTURE_2D, 0, 4,
                    gdk_pixbuf_get_width (hwin->texture),
                    gdk_pixbuf_get_height (hwin->texture),
                    0, GL_RGBA,
                    GL_UNSIGNED_BYTE,
                    gdk_pixbuf_get_pixels (hwin->texture));
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glShadeModel (GL_FLAT);
    }

  glLightfv (GL_LIGHT0, GL_POSITION, pos);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, white);
  glLightfv (GL_LIGHT0, GL_SPECULAR, white);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);

  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);
  glEnable (GL_LINE_SMOOTH);

  glEnable (GL_NORMALIZE);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  glEnable (GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(1.0, 1.0);

#if 0
  g_printerr ("\n");
  g_printerr ("GL_RENDERER   = %s\n", (char *) glGetString (GL_RENDERER));
  g_printerr ("GL_VERSION    = %s\n", (char *) glGetString (GL_VERSION));
  g_printerr ("GL_VENDOR     = %s\n", (char *) glGetString (GL_VENDOR));
  g_printerr ("GL_EXTENSIONS = %s\n", (char *) glGetString (GL_EXTENSIONS));
  g_printerr ("\n");
#endif

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/
}


GtkWidget *
helix_create_control_panel (HelixWindow *hwin)
{
  GtkWidget *win, *table;
  gint       row;

  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (win), "Helix controller");
  gtk_window_set_default_size (GTK_WINDOW (win), 500, -1);
  gtk_container_set_border_width (GTK_CONTAINER (win), 6);
  g_signal_connect (G_OBJECT (win), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  table = gtk_table_new (HELIX_SLOT_LAST, 3, FALSE);
  gtk_container_add (GTK_CONTAINER (win), table);
  gtk_table_set_row_spacings (GTK_TABLE (table), 5);
  gtk_table_set_col_spacings (GTK_TABLE (table), 5);

  for (row = 0; row < HELIX_SLOT_LAST; row++)
    {
      GtkWidget *label, *scale;
      gchar *name;

      g_object_get (hwin->adjustments[row], "name", &name, NULL);

      label = gtk_label_new (name);
      gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
      gtk_table_attach (GTK_TABLE (table),
                        label, 0, 1, row, row + 1,
                        GTK_FILL | GTK_SHRINK,
                        GTK_FILL | GTK_SHRINK,
                        0, 0);

      scale = gtk_hscale_new (hwin->adjustments[row]);
      gtk_scale_set_digits (GTK_SCALE (scale), 3);
      gtk_scale_set_value_pos (GTK_SCALE (scale), GTK_POS_RIGHT);
      gtk_table_attach_defaults (GTK_TABLE (table),
                                 scale, 1, 2, row, row + 1);

      g_free (name);
    }

  gtk_widget_show_all (table);

  return win;
}


static void
helix_setup_adjustments (HelixWindow *hwin)
{
  gint i;
  gboolean cyclic;

  hwin->midi_source = g_object_new (MIDI_TYPE_SOURCE,
                                    "device", "alsa",
                                    "channel", -1,
                                    NULL);

  for (i = 0; i < HELIX_SLOT_LAST; i++)
    {
      gdouble lower, upper, value, step, page;
      HelixSlotType slot;
      gint controller;
      gchar *name;

      slot       = helix_defaults[i].slot;

      lower      = helix_defaults[i].lower;
      upper      = helix_defaults[i].upper;
      value      = helix_defaults[i].value;

      controller = helix_defaults[i].midi_ctrl;
      cyclic     = helix_defaults[i].cyclic;
      name       = helix_defaults[i].name;

      step = (upper - lower) / 99;
      page = (upper - lower) / 11;

      if (hwin->adjustments[slot] != NULL)
        {
          GEnumClass *enum_class;
          GEnumValue *slot_value;

          enum_class = g_type_class_ref (HELIX_TYPE_SLOT_TYPE);
          slot_value = g_enum_get_value (enum_class, slot);

          g_printerr ("%s is predefined at least twice, aborting!\n",
                      slot_value->value_name);
          if (slot == 0)
            g_printerr ("maybe not enough elements in helix_defaults?\n");

          exit (1);
        }

      hwin->adjustments[slot] = g_object_new (MIDI_TYPE_ADJUSTMENT,
                                              "lower", lower,
                                              "upper", upper,
                                              "value", value,
                                              "target-value", value,
                                              "step-increment", step,
                                              "page-increment", page,
                                              "midi-source", hwin->midi_source,
                                              "is-cyclic", cyclic,
                                              "controller", controller,
                                              "name", name,
                                              NULL);
    }
}


HelixWindow *
helix_setup_ui ()
{
  GdkGLConfig *glconfig;
  GtkWidget *vbox;
  HelixWindow *hwin;

  hwin = g_new0 (HelixWindow, 1);

  hwin->texture = gdk_pixbuf_new_from_file ("texture.png", NULL);

  if (hwin->texture)
    {
      gboolean success = TRUE;

      success = success && gdk_pixbuf_get_has_alpha (hwin->texture);
      success = success && (gdk_pixbuf_get_n_channels (hwin->texture) == 4);
      success = success && (gdk_pixbuf_get_bits_per_sample (hwin->texture) == 8);

      success = success && (gdk_pixbuf_get_width (hwin->texture) *
                               gdk_pixbuf_get_n_channels (hwin->texture)
                            == gdk_pixbuf_get_rowstride (hwin->texture));

      if (!success)
        {
          g_object_unref (G_OBJECT (hwin->texture));
          hwin->texture = NULL;
        }
    }

  /* create timer */
  if (hwin->timer == NULL)
    hwin->timer = g_timer_new ();

  g_timer_start (hwin->timer);

  hwin->fps = 50;
  hwin->nelements = 54;
  hwin->elements = g_new0 (HelixElement, hwin->nelements);
  hwin->master_element = g_new0 (HelixElement, 1);

  helix_elements_configure (hwin);

  /*
   * Configure OpenGL-capable visual.
   */

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL)
    {
      g_printerr ("*** Cannot find the double-buffered visual.\n");
      g_printerr ("*** Trying single-buffered visual.\n");

      /* Try single-buffered visual */
      glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB   |
                                            GDK_GL_MODE_DEPTH);
      if (glconfig == NULL)
        {
          g_printerr ("*** No appropriate OpenGL-capable visual found.\n");
          exit (1);
        }
    }

  /*
   * Top-level window.
   */

  hwin->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (hwin->window), "Helix");

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (hwin->window), TRUE);

  g_signal_connect (G_OBJECT (hwin->window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  /*
   * VBox.
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (hwin->window), vbox);
  gtk_widget_show (vbox);

  /*
   * Drawing area for drawing OpenGL scene.
   */

  hwin->drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (hwin->drawing_area, 600, 600);

  /* Set OpenGL-capability to the widget. */
  gtk_widget_set_gl_capability (hwin->drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (hwin->drawing_area,
                         GDK_VISIBILITY_NOTIFY_MASK);

  g_signal_connect_after (G_OBJECT (hwin->drawing_area), "realize",
                          G_CALLBACK (helix_init_gl), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "configure_event",
                    G_CALLBACK (helix_configure_event), NULL);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "expose_event",
                    G_CALLBACK (helix_frame_draw_stereo), hwin);

  g_signal_connect (G_OBJECT (hwin->drawing_area), "map_event",
                    G_CALLBACK (helix_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "unmap_event",
                    G_CALLBACK (helix_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "visibility_notify_event",
                    G_CALLBACK (helix_visibility_update), hwin);

  g_signal_connect (G_OBJECT (hwin->window), "key_press_event",
                    G_CALLBACK (helix_key_press_event), hwin);

  gtk_box_pack_start (GTK_BOX (vbox), hwin->drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (hwin->drawing_area);

  /*
   * Show window.
   */

  gtk_widget_show (hwin->window);

  helix_setup_adjustments (hwin);

  /*
   * Timeout function to maintain framerate
   */
  hwin->timeout_id = g_timeout_add_full (GDK_PRIORITY_REDRAW,
                                         1000 / hwin->fps,
                                         (GSourceFunc) helix_frame_next,
                                         hwin,
                                         NULL);
  return hwin;
}


int
main (int   argc,
      char *argv[])
{
  HelixWindow     *hwin;

  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);

  hwin = helix_setup_ui ();

  gtk_main ();

  return 0;
}
