#include <glib-object.h>
#include <math.h>

#include "midi-source.h"
#include "midi-adjustment.h"

enum
{
  PROP_0,
  PROP_NAME,
  PROP_MIDI_SOURCE,
  PROP_CONTROLLER,
  PROP_IS_CYCLIC,
  PROP_TARGET_VALUE,
};


static void     midi_adjustment_class_init      (MidiAdjustmentClass  *klass);

static void     midi_adjustment_finalize        (GObject         *object);
static void     midi_adjustment_set_property    (GObject         *object,
                                                 guint            property_id,
                                                 const GValue    *value,
                                                 GParamSpec      *pspec);
static void     midi_adjustment_get_property    (GObject         *object,
                                                 guint            property_id,
                                                 GValue          *value,
                                                 GParamSpec      *pspec);

static void     midi_adjustment_set_midi_source (MidiAdjustment *midi_adj,
                                                 MidiSource     *midi_source);
static void     midi_adjustment_process_event   (MidiAdjustment *midi_adj,
                                                 MidiEvent      *event,
                                                 gpointer        data);
static gboolean midi_adjustment_update_value    (MidiAdjustment *midi_adj);

static GObjectClass *parent_class = NULL;


GType
midi_adjustment_get_type (void)
{
  static GType object_type = 0;

  if (! object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (MidiAdjustmentClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) midi_adjustment_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data     */
        sizeof (MidiAdjustment),
        0,              /* n_preallocs    */
        (GInstanceInitFunc) NULL,
      };

      object_type = g_type_register_static (GTK_TYPE_ADJUSTMENT,
                                            "MidiAdjustment",
                                            &object_info, 0);
    }

  return object_type;
}


static void
midi_adjustment_class_init (MidiAdjustmentClass *klass)
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize     = midi_adjustment_finalize;
  object_class->set_property = midi_adjustment_set_property;
  object_class->get_property = midi_adjustment_get_property;

  g_object_class_install_property (object_class,
                                   PROP_NAME,
                                   g_param_spec_string ("name",
                                                        NULL, NULL,
                                                        NULL,
                                                        G_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT));

  g_object_class_install_property (object_class,
                                   PROP_MIDI_SOURCE,
                                   g_param_spec_object ("midi_source",
                                                        NULL, NULL,
                                                        MIDI_TYPE_SOURCE,
                                                        G_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT));

  g_object_class_install_property (object_class,
                                   PROP_CONTROLLER,
                                   g_param_spec_int ("controller",
                                                     NULL, NULL,
                                                     -1, 119, -1,
                                                     G_PARAM_READWRITE |
                                                     G_PARAM_CONSTRUCT));

  g_object_class_install_property (object_class,
                                   PROP_IS_CYCLIC,
                                   g_param_spec_boolean ("is_cyclic",
                                                         NULL, NULL,
                                                         FALSE,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT));

  g_object_class_install_property (object_class,
                                   PROP_TARGET_VALUE,
                                   g_param_spec_double ("target-value",
                                                        NULL, NULL,
                                                        -G_MAXDOUBLE,
                                                        G_MAXDOUBLE, 0.0,
                                                        G_PARAM_READWRITE));
}


static void
midi_adjustment_finalize (GObject *object)
{
  MidiAdjustment *midi_adj = MIDI_ADJUSTMENT (object);

  midi_adjustment_set_midi_source (midi_adj, NULL);

  if (midi_adj->name)
    g_free (midi_adj->name);

  if (!midi_adj->update_handler)
    gtk_timeout_remove (midi_adj->update_handler);
  midi_adj->update_handler = 0;

  G_OBJECT_CLASS (parent_class)->finalize (object);
}


static void
midi_adjustment_set_property (GObject      *object,
                              guint         property_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
  MidiAdjustment *midi_adj = MIDI_ADJUSTMENT (object);
  GtkAdjustment  *adj = GTK_ADJUSTMENT (object);
  gdouble         new_value;

  switch (property_id)
    {
    case PROP_NAME:
      if (midi_adj->name)
        g_free (midi_adj->name);
      midi_adj->name = g_value_dup_string (value);
      break;

    case PROP_MIDI_SOURCE:
      midi_adjustment_set_midi_source (midi_adj,
                                       MIDI_SOURCE (g_value_get_object (value)));
      break;

    case PROP_CONTROLLER:
      midi_adj->controller = g_value_get_int (value);
      break;

    case PROP_IS_CYCLIC:
      midi_adj->is_cyclic = g_value_get_boolean (value);
      break;

    case PROP_TARGET_VALUE:
      new_value = g_value_get_double (value);
      if (midi_adj->is_cyclic)
        {
          while (new_value > adj->upper)
            new_value -= (adj->upper - adj->lower);
          while (new_value < adj->lower)
            new_value += (adj->upper - adj->lower);
        }
      midi_adj->target_value = CLAMP (new_value, adj->lower, adj->upper);
      if (!midi_adj->update_handler)
        midi_adj->update_handler = gtk_timeout_add (20, (GtkFunction) midi_adjustment_update_value, midi_adj);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}


static void
midi_adjustment_get_property (GObject    *object,
                              guint       property_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
  MidiAdjustment *midi_adj = MIDI_ADJUSTMENT (object);

  switch (property_id)
    {
    case PROP_NAME:
      g_value_set_string (value, midi_adj->name);
      break;
    case PROP_MIDI_SOURCE:
      g_value_set_object (value, midi_adj->midi_source);
      break;
    case PROP_CONTROLLER:
      g_value_set_int (value, midi_adj->controller);
      break;
    case PROP_IS_CYCLIC:
      g_value_set_boolean (value, midi_adj->is_cyclic);
      break;
    case PROP_TARGET_VALUE:
      g_value_set_double (value, midi_adj->target_value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}


static void
midi_adjustment_set_midi_source (MidiAdjustment *midi_adj,
                                 MidiSource     *midi_source)
{
  g_return_if_fail (midi_source == NULL || MIDI_IS_SOURCE (midi_source));

  if (midi_adj->midi_source)
    {
      g_signal_handler_disconnect (midi_adj->midi_source, midi_adj->midi_handler);
      midi_adj->midi_handler = 0;
      g_object_unref (midi_adj->midi_source);
      midi_adj->midi_source = NULL;
    }

  if (midi_source)
    {
      midi_adj->midi_source = midi_source;
      g_object_ref (midi_source);
      midi_adj->midi_handler =
          g_signal_connect_swapped (midi_source, "midi-event",
                                    G_CALLBACK (midi_adjustment_process_event),
                                    midi_adj);
    }
}


static void
midi_adjustment_process_event (MidiAdjustment *midi_adj,
                               MidiEvent      *event,
                               gpointer        data)
{
  GtkAdjustment *adj = GTK_ADJUSTMENT (midi_adj);

  if (event->type == MIDI_EVENT_CONTROLLER &&
      (midi_adj->controller == -1 ||
       midi_adj->controller == event->note))
    {
      gdouble new_val;

      new_val = adj->lower + event->value * (adj->upper - adj->lower);

      g_object_set (G_OBJECT (midi_adj), "target-value", new_val, NULL);
    }
}

#define MAX_ACCEL 0.005
#define MAX_SPEED 0.1

#define SLOWING_DISTANCE 0.5

static gboolean
midi_adjustment_update_value (MidiAdjustment *midi_adj)
{
  GtkAdjustment *adj = GTK_ADJUSTMENT (midi_adj);
  gdouble distance, speed, offset, steering, new_val;

  g_return_val_if_fail (MIDI_IS_ADJUSTMENT (midi_adj), FALSE);

  /* modeled as on http://www.red3d.com/cwr/steer/gdc99/ */

  /* Arrival behaviour:
   *
   *     target_offset = target - position
   *     distance = length (target_offset)
   *     ramped_speed = max_speed * (distance / slowing_distance)
   *     clipped_speed = minimum (ramped_speed, max_speed)
   *     desired_velocity = (clipped_speed / distance) * target_offset
   *     steering = desired_velocity - velocity
   */

  /* offset is normed into the [0,1] interval. */
  offset = (midi_adj->target_value - adj->value) / (adj->upper - adj->lower);
  distance = fabs (offset);

  if (midi_adj->is_cyclic && distance >= 0.5)
    {
      if (offset > 0)
        offset = offset - 1;
      else
        offset = offset + 1;

      distance = 1 - distance;
    }

  if (distance < 0.00001)
    {
      midi_adj->update_handler = 0;
      gtk_adjustment_set_value (adj, midi_adj->target_value);
      return FALSE;
    }

  steering = MIN (MAX_SPEED * distance / SLOWING_DISTANCE,
                  MAX_SPEED) / distance * offset - midi_adj->speed;

  speed = midi_adj->speed + CLAMP (steering, -MAX_ACCEL, MAX_ACCEL);
  midi_adj->speed = CLAMP (speed, -MAX_SPEED, MAX_SPEED);

  /* map back to original range */
  speed = speed * (adj->upper - adj->lower);

  new_val = adj->value + speed;

  if (midi_adj->is_cyclic)
    {
      while (new_val > adj->upper)
        new_val -= (adj->upper - adj->lower);
      while (new_val < adj->lower)
        new_val += (adj->upper - adj->lower);
    }

  gtk_adjustment_set_value (adj, new_val);

  return TRUE;
}
